// vim:syntax=lpc
//! @ignore
inherit Public.System@module@;

constant __author = "Arne Goedeke <pike@laramies.com>";
constant __version = "0.1";
constant __components = ({ "Public.pmod/System.pmod/Inotify.pmod/module.pmod", "Public.pmod/System.pmod/___Inotify.so" });
//! @endignore


string describe_mask(int mask) {
	array(string) list = ({});
	
	foreach (indices(this_program);; string name) {
		if (has_prefix(name, "IN_")) {
			int value = `[](this_program, name);

			if (value & mask) list += ({ name });
		}
	}

	return list * "|";
}

//! More convenient interface to inotify(7). Automatically reads events from the inotify file descriptor and parses them.
class Instance {
	// I dont normally use private. In this case its possible to break things by messing with file
	// and the inotify fd.
	private object instance;
	private Stdio.File file;

	class Watch {
		string name;
		function callback;
		int mask;
		array extra;

		void create(string name, int mask, function callback, array extra) {
			this_program::name = name;
			this_program::mask = mask;
			this_program::callback = callback;
			this_program::extra = extra;
		}
	}

	class DirWatch {
		inherit Watch;

		void handle_event(int wd, int mask, int cookie, int|string name) {
			if (name) {
				name = (has_suffix(this_program::name, "/") ? this_program::name : (this_program::name+"/")) + name;	
			} else {
				name = this_program::name;
			}

			// maybe its good to use a call_out here, but it will mess up the order of the events. until
			// there is an ordered call_out(.., 0,..) we will do this.
			callback(mask, cookie, name, @extra);
		}
	}

	class FileWatch {
		inherit Watch;
		
		void handle_event(int wd, int mask, int cookie, int|string name) {
			callback(mask, cookie, this_program::name, @extra);
		}
	}

	mapping(int:object) watches = ([]);

	void parse(mixed id, string data) {
		while (sizeof(data)) {
			array a;
			mixed err = catch {
				a = parse_event(data);
			};

			if (err) {
				// TODO: might have a partial even struct here which gets completed
				// by the next call?? maybe add an internal buffer.
				werror("Could not parse inotify event: %s\n", describe_error(err));
				return;
			}

			object watch;

			if (watch = watches[a[0]]) {
				watch->handle_event(a[0], a[1], a[2], a[3]);
			} else {
				werror("No callback for event %O\n", a);
			}

			data = data[a[4]..];
		}
	}

	void create() {
		instance = _Instance();
		file = Stdio.File(instance->get_fd(), "r");
		file->set_nonblocking();
		file->set_read_callback(parse);
	}

	//! Add a watch for a certain file and a set of events specified by @expr{mask@}. The function 
	//! @expr{callback@} will be called when such an event occurs. The arguments to the callback
	//! will be the events mask, the cookie, the @expr{filename@} and @expr{extra@}.
	//! @returns
	//! 	Returns a watch descriptor which can be used to remove the watch.
	//! @note
	//! 	When adding a second watch for the same file the old one will be removed
	//! 	unless @[Public.System.Inotify.IN_MASK_ADD] is contained in @expr{mask@}.
	//! @note
	//! 	The mask of an event may be a subset of the mask given when adding the watch.
	//! @note
	//! 	In case the watch is added for a regular file, the filename will be passed
	//! 	to the callback. In case the watch is added for a directory, the name of the file
	//! 	to which the event happened inside the directory will be concatenated.
	int add_watch(string filename, int mask, function callback, mixed ... extra) {
		int wd = instance->add_watch(filename, mask);

#if constant(Public.System@module@.IN_MASK_ADD)
		// just in case we want to use the mask in the watches later,
		// they just be correct
		if (mask & IN_MASK_ADD && has_index(watches, wd)) {
			mask |= watches[wd]->mask;
		}
#endif

		if (Stdio.is_dir(filename)) {
			watches[wd] = DirWatch(filename, mask, callback, extra);
		} else {
			watches[wd] = FileWatch(filename, mask, callback, extra);
		}

		return wd;
	}

	//! Remove the watch associated with the watch descriptor @expr{wd@}.
	void rm_watch(int wd) {
		if (!has_index(watches, wd)) {
			error("There is no watch with id %d here.\n", wd);
		}
		
		m_delete(watches, wd);
		instance->rm_watch(wd);	
	}
}
